import numpy as np
from nltk import tokenize
from keras.utils import to_categorical
from keras import backend as K
from keras.preprocessing.text import Tokenizer
from sklearn.preprocessing import LabelBinarizer
from keras.preprocessing.sequence import pad_sequences
from sklearn.model_selection import train_test_split


def f1(y_true, y_pred):
    def recall(y_true, y_pred):
        true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
        recall = true_positives / (possible_positives + K.epsilon())
        return recall

    def precision(y_true, y_pred):
        true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
        precision = true_positives / (predicted_positives + K.epsilon())
        return precision
    precision = precision(y_true, y_pred)
    recall = recall(y_true, y_pred)
    return 2*((precision*recall)/(precision+recall+K.epsilon()))


def readfile(filename):
    filecontents = list()
    with open(filename) as fh:
        doccontent = list()
        for line in fh:
            if len(line) == 1:
                filecontents.append(doccontent)
                doccontent = list()
            else:
                splitline = line.split()
                if len(splitline) > 1:
                    doccontent.append((splitline[0], splitline[1]))
    return [x for x in filecontents if x != []]


def extract_labels(fullset):
    rawlabels = list()
    for document in fullset:
        entries = list()
        for entry in document:
            entries.append(entry[1])
        rawlabels.append(entries)
    return rawlabels


def get_documents(fullset):
    docs = list()
    for document in fullset:
        entries = list()
        for entry in document:
            entries.append(entry[0])
        docs.append(entries)
    return docs


def flat_words(fullset):
    allwords = list()
    for document in fullset:
        for entry in document:
            allwords.append(entry[0])
    return allwords


def sentence_splitter(documents):
    documents_split = []
    for doc in documents:
        full_doc = " ".join(doc)
        sent_doc = tokenize.sent_tokenize(full_doc)
        combined_doc = []
        for sent in sent_doc:
            split_sent = sent.split()
            combined_doc.append(split_sent)
        documents_split.append(combined_doc)
    return documents_split


def build_encoding(documents, labels):
    words, tags = set([]), set([])

    for s in documents:
        for w in s:
            words.add(w.lower())

    for ts in labels:
        for t in ts:
            tags.add(t)

    word2index = {w: i + 2 for i, w in enumerate(sorted(list(words)))}
    word2index['-PAD-'] = 0
    word2index['-OOV-'] = 1

    tag2index = {t: i + 1 for i, t in enumerate(sorted(list(tags)))}
    tag2index['-PAD-'] = 0

    return (word2index, tag2index)


def encode_data(documents, labels, word2index, tag2index):
    documents_X, labels_Y = [], []
    for i, doc in enumerate(documents):
        doc_int, lab_int = [], []
        index = 0
        for s in doc:
            s_int, l_int = [], []
            for w in s:
                if len(labels[i]) > index:
                    l_int.append(tag2index[labels[i][index]])
                index += 1
                try:
                    s_int.append(word2index[w.lower()])
                except KeyError:
                    s_int.append(word2index['-OOV-'])
            doc_int.append(s_int)
            lab_int.append(l_int)
        documents_X.append(doc_int)
        labels_Y.append(lab_int)

    return (documents_X, labels_Y)


def scorer(goldlabels, predictedlabels, B=0, I=1):
    truepos = 0
    falsepos = 0
    falseneg = 0
    for i, label in enumerate(predictedlabels):
        if label == B and goldlabels[i] == B or label == I and goldlabels[i] == I:
            truepos += 1
        elif label == B or label == I:
            falsepos += 1
        elif goldlabels[i] == B or goldlabels[i] == I:
            falseneg += 1
    try:
        precision = truepos / (truepos + falsepos)
        recall = truepos / (truepos + falseneg)
    except ZeroDivisionError:
        precision = 0.00001
        recall = 0.00001

    try:
        f1 = 2 * (precision * recall) / (precision + recall)
    except ZeroDivisionError:
        f1 = 0

    print('true positives:', truepos)
    print('false positives:', falsepos)
    print('false negatives:', falseneg)

    print('Precision:', precision)
    print('Recall:', recall)
    print('F1:', f1)


def writepredictions(flatwords, predictedlabels, filename):
    with open(filename, 'w') as fh:
        for i, word in enumerate(flatwords):
            fh.write(word + "\t" + predictedlabels[i] + '\n')


def seqnn_processing(train_d, test_d, train_l, test_l, word2index, label2index):
    n_tags = len(label2index)

    train_d_p = sentence_splitter(train_d)
    test_d_p = sentence_splitter(test_d)

    train_X, train_y = encode_data(train_d_p, train_l, word2index, label2index)
    test_X, test_y = encode_data(test_d_p, test_l, word2index, label2index)

    train_X = [item for sublist in train_X for item in sublist]
    test_X = [item for sublist in test_X for item in sublist]

    train_y = [item for sublist in train_y for item in sublist]
    test_y = [item for sublist in test_y for item in sublist]

    max_len = max(len(max(train_X, key=len)), len(max(test_X, key=len)))

    train_X = pad_sequences(train_X, maxlen=max_len, padding='post')
    test_X = pad_sequences(test_X, maxlen=max_len, padding='post')
    train_y = pad_sequences(train_y, maxlen=max_len, padding='post')
    test_y = pad_sequences(test_y, maxlen=max_len, padding='post')

    train_y = np.asarray([to_categorical(i, num_classes=n_tags) for i in train_y])
    test_y = np.asarray([to_categorical(i, num_classes=n_tags) for i in test_y])

    return (train_X, test_X, train_y, test_y)


def fnn_processing(train_d, test_d, train_l, test_l, n_words):
    flat_train_docs = [item for sublist in train_d for item in sublist]
    flat_test_docs = [item for sublist in test_d for item in sublist]

    tokenizer = Tokenizer(num_words=n_words//2)
    tokenizer.fit_on_texts(flat_train_docs + flat_test_docs)

    train_X = tokenizer.texts_to_matrix(flat_train_docs, mode='tfidf')
    test_X = tokenizer.texts_to_matrix(flat_test_docs, mode='tfidf')

    flat_train_labels = [item for sublist in train_l for item in sublist]
    flat_test_labels = [item for sublist in test_l for item in sublist]
    encoder = LabelBinarizer()
    encoder.fit(flat_train_labels)
    train_y = encoder.transform(flat_train_labels)
    test_y = encoder.transform(flat_test_labels)

    return (train_X, test_X, train_y, test_y)


def process_data(inputfile, slice_size=-1):
    raw_data = readfile(inputfile)

    labels = extract_labels(raw_data)
    documents = get_documents(raw_data)

    labels = labels[:slice_size]
    documents = documents[:slice_size]
    word2index, label2index = build_encoding(documents, labels)

    train_docs, test_docs, train_labels, test_labels = train_test_split(documents, labels, test_size=0.1)
    return (train_docs, test_docs, train_labels, test_labels, word2index, label2index)
